/*
    Copyright (C) 2003 Evan Buswell

    This file borrows heavily from:
    mpg321 - ``a fully free clone of mpg123''.
    Copyright (C) 2001 Joe Drew

    I signed his long / important comments with his initials.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <mad.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef REAL_AO
# include "ao.h"
#else
# include "fakeao.h"
#endif

#include "options.h"

#ifdef HAVE_CONFIG_H
# include "config.h" /* need this for big / little endian */
#endif

/* The following two routines and data structure are from the ever-brilliant
   Rob Leslie. --JD
*/

struct audio_dither {
    mad_fixed_t error[3];
    mad_fixed_t random;
};

/*
* NAME:        prng()
* DESCRIPTION: 32-bit pseudo-random number generator --JD
*/
static inline
unsigned long prng(unsigned long state) {
    return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

/*
* NAME:        audio_linear_dither()
* DESCRIPTION: generic linear sample quantize and dither routine --JD
*/
inline
signed long audio_linear_dither(unsigned int bits, mad_fixed_t sample,
                                struct audio_dither *dither) {
    unsigned int scalebits;
    mad_fixed_t output, mask, random;
    
    enum {
	MIN = -MAD_F_ONE,
	MAX =  MAD_F_ONE - 1
    };
    
    /* noise shape */
    sample += dither->error[0] - dither->error[1] + dither->error[2];
    
    dither->error[2] = dither->error[1];
    dither->error[1] = dither->error[0] / 2;
    
    /* bias */
    output = sample + (1L << (MAD_F_FRACBITS + 1 - bits - 1));
    
    scalebits = MAD_F_FRACBITS + 1 - bits;
    mask = (1L << scalebits) - 1;
    
    /* dither */
    random  = prng(dither->random);
    output += (random & mask) - (dither->random & mask);
    
    dither->random = random;
    
    /* clip */
    if (output > MAX) {
	output = MAX;
	
	if (sample > MAX)
	    sample = MAX;
    }
    else if (output < MIN) {
	output = MIN;
	
	if (sample < MIN)
	    sample = MIN;
    }
    
    /* quantize */
    output &= ~mask;
    
    /* error feedback */
    dither->error[0] = sample - output;
    
    /* scale */
    return output >> scalebits;
}

#undef OLDOUTPUT
#ifdef OLDOUTPUT /* the old function... */
int output(struct mad_header const *header,
	   struct mad_pcm *pcm) {
    register int nsamples = pcm->length;
    mad_fixed_t const *left_ch = pcm->samples[0], *right_ch = pcm->samples[1];
    
    static unsigned char stream[1152*4]; /* 1152 because that's what mad has as a max; *4 because
                                    there are 4 distinct bytes per sample (in 2 channel case)
				    --JD*/
    static unsigned int rate = 0;
    static int channels = 0;
    static struct audio_dither dither;
    
    register char * ptr = stream;
    register signed int sample;
    register mad_fixed_t tempsample;
    
    /* We need to know information about the file before we can open the playdevice
       in some cases. So, we do it here. --JD*/
    if (!playdevice) {
        channels = MAD_NCHANNELS(header);
        rate = header->samplerate;
        open_ao_playdevice(header);
    } else if ((channels != MAD_NCHANNELS(header) || rate != header->samplerate) && playdevice_is_live()) {
#ifdef REAL_AO
        ao_close(playdevice);
#endif
        channels = MAD_NCHANNELS(header);
        rate = header->samplerate;
        open_ao_playdevice(header);
    }
    
    if (pcm->channels == 2) {
        while (nsamples--) {
            tempsample = (mad_fixed_t)((*left_ch++ * (double)options.volume)/MAD_F_ONE);
            sample = (signed int) audio_linear_dither(16, tempsample, &dither);
	    
#ifndef WORDS_BIGENDIAN
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
#else
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
#endif

            tempsample = (mad_fixed_t)((*right_ch++ * (double)options.volume)/MAD_F_ONE);
            sample = (signed int) audio_linear_dither(16, tempsample, &dither);
#ifndef WORDS_BIGENDIAN
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
#else
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
#endif
        }

        ao_play(playdevice, stream, pcm->length * 4);
    } else if (options.opt & M_FORCE_STEREO) {
        while (nsamples--) {
            tempsample = (mad_fixed_t)((*left_ch++ * (double)options.volume)/MAD_F_ONE);
            sample = (signed int) audio_linear_dither(16, tempsample, &dither);
	    
            /* Just duplicate the sample across both channels. */
#ifndef WORDS_BIGENDIAN
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
#else
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
#endif
        }
	
        ao_play(playdevice, stream, pcm->length * 4);
    } else /* Just straight mono output */ {
        while (nsamples--) {
            tempsample = (mad_fixed_t)((*left_ch++ * (double)options.volume)/MAD_F_ONE);
            sample = (signed int) audio_linear_dither(16, tempsample, &dither);
            
#ifndef WORDS_BIGENDIAN
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
#else
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
#endif
        }
        ao_play(playdevice, stream, pcm->length * 2);
    }

    return(0);
}
#endif /* OLDOUTPUT */

int output2(struct mad_header const *header,
	    struct mad_pcm *pcm) {
    signed int sample;
    int i;
    int offset;
    int so;
    unsigned char buffer[1152];
    
    static struct audio_dither ldither;
    static struct audio_dither rdither;
    
    static unsigned int rate = 0;
    static int channels = 0;
    
    if (!playdevice) {
	channels = MAD_NCHANNELS(header);
	rate = header->samplerate;
	open_ao_playdevice(header);
    } else if ((channels != MAD_NCHANNELS(header) ||
		rate != header->samplerate) && playdevice_is_live()) {
#ifdef REAL_AO
	ao_close(playdevice);
#endif
	channels = MAD_NCHANNELS(header);
	rate = header->samplerate;
	open_ao_playdevice(header);
    }
    
    offset = 0;
    for(i = 0; i < pcm->length; i++) {
	if((offset + 4) >= 1152) {
	    ao_play(playdevice, buffer, offset);
	    offset = 0;
	}
	/* left channel */
	sample = (signed int) audio_linear_dither(16, pcm->samples[0][i], &ldither);
	buffer[offset] = (unsigned char) (sample >> 0);
	offset++;
	buffer[offset] = (unsigned char) (sample >> 8);
	offset++;
	/* right channel */
	if(MAD_NCHANNELS(header) == 2)
	    sample = (signed int) audio_linear_dither(16, pcm->samples[1][i], &ldither);
	buffer[offset] = (unsigned char) (sample >> 0);
	offset++;
	buffer[offset] = (unsigned char) (sample >> 8);
	offset++;
    }
    ao_play(playdevice, buffer, offset);
}

int decode_frame(unsigned char *buffer, unsigned long length,
		 unsigned long offset, int skipped /* did we drop some frames? */) {
    static struct mad_stream stream;
    static int firstrun = 1;
    static int errorlast = 0;
    static struct mad_frame frame;
    static struct mad_synth synth;
    
#ifdef DEBUG
    fprintf(stderr, "Decoder called.\n");
#endif
    
    if(firstrun || skipped || errorlast) {
	if(!firstrun) {
	    /* kill old structures if they exist */
	    mad_synth_finish(&synth);
	    mad_stream_finish(&stream);
	    mad_frame_finish(&frame);
	}
	/* initialize all structures... */
	mad_synth_init(&synth);
	mad_stream_init(&stream);
	mad_frame_init(&frame);
	
	/* stream.options = 0; here's where to set stream options if we need them...we don't. */
	
	mad_stream_buffer(&stream, buffer, length);
	mad_stream_skip(&stream, offset);
	firstrun = 0;
    } else {
	/* restore values */
	stream.next_frame = buffer + offset;
	stream.sync = 1;
    }

    stream.error = 0;
  
    /* decode it */
    if(mad_frame_decode(&frame, &stream) == -1) {
	fprintf(stderr, "mad_frame_decode failed: ");
	fprintf(stderr, "%s\n", mad_stream_errorstr(&stream));
	if(!MAD_RECOVERABLE(stream.error)) {
	    fprintf(stderr, "Skipping this frame.\n");
	    errorlast = 1;
	    return(-1);
	}
    }
#ifdef DEBUG
    fprintf(stderr, "Preparing to synth the frame...\n"); /* see where we are when we crash. */
#endif
    mad_synth_frame(&synth, &frame);
    
    output2(&frame.header, &synth.pcm);

    return(0);
}
