/*
    Copyright (C) 2003 Evan Buswell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <asm/unistd.h>
#include <mad.h>
#include <pthread.h>

#include "net.h"
#include "../common/frame.h"
#include "../common/port.h"


pthread_mutex_t decoder_is_running = PTHREAD_MUTEX_INITIALIZER;

struct sockaddr_in net_raddr;
int net_raddr_len = sizeof(struct sockaddr_in);

int netinit(char *hostname, int port) {
    struct hostent *host;
    int s;
    int on = 1;
    struct sockaddr_in laddr;

    /* bind */
    if((s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
	perror("Failed to create socket");
	return(-1);
    }
    if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(int)) == -1) {
	perror("setsockopt");
	return(-1);
    }
    laddr.sin_family = AF_INET;
    laddr.sin_port = htons(BROADCASTPORT_C);
    laddr.sin_addr.s_addr = INADDR_ANY;
    if(bind(s, (struct sockaddr *) &laddr, sizeof(struct sockaddr_in)) == -1) {
	fprintf(stderr, "Failed to bind to port %d\n", port);
	perror("");
    }

    /* remote */
    if((host = gethostbyname(hostname)) == NULL) {
	fprintf(stderr, "Failed to resolve hostname %s\n", hostname);
	return(-1);
    }
    net_raddr.sin_family = AF_INET;
    net_raddr.sin_port = htons(port);
    net_raddr.sin_addr.s_addr = *((u_int32_t *) host->h_addr);
    return(s);
}

struct dfwvars {
    unsigned char *buffer;
    unsigned long length;
    unsigned long offset;
    int skipped;
};

void *decode_frame_wrapper(void *arg) {
    struct dfwvars *vars;
    vars = (struct dfwvars *) arg;
    pthread_mutex_lock(&decoder_is_running);
    decode_frame(vars->buffer, vars->length, vars->offset, vars->skipped);
    pthread_mutex_unlock(&decoder_is_running);
}

/* get_data(int s)
 * Parameters: s is a udp socket bound to the local listening address.
 * Variables: net_raddr: the address to connect to,
 *            net_raddr_len: the length of net_raddr structure.
 * Never returns.
 */
void get_data(int s) {
    /* flags: */
    int run_decoder = 0;
    int lostsync = 0;
    int firstrun = 1;

    /* buffers: */
    unsigned char *buffer[BUFFER_DEPTH + 1];
    /* The array is used as follows:
       [0] = the beginning of the buffer
       [BUFFER_END ( == BUFFER_DEPTH )] = the last valid byte.
       [1] through [BUFFEREND - 1] = pointers to frames within the buffer.
    */
    int lengthleft; /* the space left in the buffer */
    int cur_depth = 0; /* the frame we are currently working on */
    int offset = 0; /* an offset into the current frame */
    
    int i; /* all purpose temporary integer */

    pthread_t d_thread = 0;
    
    struct mad_header header;
    struct mad_stream stream;
    
    buffer[0] = (unsigned char *) malloc(BUFFER_SIZE);
    
    lengthleft = BUFFER_SIZE;
    buffer[BUFFER_END] = buffer[0];

    while(1) {
	if(buffer[cur_depth] + offset + MAD_BUFFER_GUARD >= buffer[BUFFER_END]
	   && !run_decoder) {
	    /* we have nothing to do except get more data.  Set socket as blocking. */
	    fcntl(s, F_SETFL, 0);
	}
	else {
	    fcntl(s, F_SETFL, O_NONBLOCK);
	}
	
	/* get more data if there's any available */
	i = recvfrom(s, buffer[BUFFER_END], lengthleft, 0,
		     (struct sockaddr *) &net_raddr, &net_raddr_len);
	if(i == -1 && errno != EAGAIN) {
	    perror("Socket error; connection lost");
	    exit(1);
	}
	if(i > 0) {
	    lengthleft -= i;
	    buffer[BUFFER_END] += i;
#ifdef DEBUG
	    fprintf(stderr, "received %d bytes\n", i);
#endif
	}

	/* grep what we have for the separator */
	while(buffer[cur_depth] + offset + MAD_BUFFER_GUARD < buffer[BUFFER_END]) {
	    if(testboundary(buffer[cur_depth], offset)) {
		mad_header_init(&header);
		mad_stream_init(&stream);
		i = decode_next_offset(buffer[cur_depth],
				       buffer[BUFFER_END] - buffer[cur_depth] /*"length"*/, 
				       &header, &stream, offset);
		mad_header_finish(&header);
		mad_stream_finish(&stream);
		if(i == -1) {
		    fprintf(stderr, "Error decoding frame: %s\n", mad_stream_errorstr(&stream));
		    lostsync = 1;
		    offset++;
		    continue;
		}
#ifdef DEBUG
		fprintf(stderr, "Boundary discovered at %d\n", offset);
		fprintf(stderr, "Frame is %d bytes long\n", i);
#endif

		
		/* offset is the beginning of a frame. */
		if(cur_depth + 1 == BUFFER_DEPTH) {
		    memcpy(buffer[0], buffer[cur_depth] + offset,
			   buffer[BUFFER_END] - (buffer[cur_depth] + offset));
		    buffer[BUFFER_END] = buffer[0] +
			(buffer[BUFFER_END] - (buffer[cur_depth] + offset));
		    lengthleft = BUFFER_SIZE - (buffer[BUFFER_END] - buffer[0]);
		    cur_depth = 0;
		} else {
		    buffer[cur_depth + 1] = buffer[cur_depth] + offset;
		    cur_depth++;
		}

		offset = i; /* the next will start at least here. */
		if(run_decoder == 1) {
		    /* if the decoder is lagging... */
		    lostsync = 1;
		}
		run_decoder = 1 ^ firstrun;
		firstrun = 0;
	    } else {
		lostsync = 1;
		offset++;
	    }
	} /* done decoding headers and such */
	
	/* run the decoder if the necessary data is present. */
	if(run_decoder && pthread_mutex_trylock(&decoder_is_running)) {
	    pthread_mutex_unlock(&decoder_is_running);
	    if(d_thread != 0) {
		pthread_join(d_thread, NULL);
	    }
#ifdef DEBUG
	    fprintf(stderr, "Trying to run decoder.\n");
#endif
	    run_decoder = 0;
	    /* we decode 1 before the current */
	    {
		int todecode;
		struct dfwvars decode_frame_vars;
		if(cur_depth == 0)
		    todecode = BUFFER_DEPTH - 1;
		else
		    todecode = cur_depth - 1;
		decode_frame_vars.buffer = buffer[0];
		decode_frame_vars.skipped = lostsync;
		decode_frame_vars.offset = buffer[todecode] - buffer[0];
		decode_frame_vars.length = BUFFER_SIZE;
		if(lostsync) {
		    fprintf(stderr, "Sync lost\n");
		}
		pthread_create(&d_thread, NULL, decode_frame_wrapper, &decode_frame_vars);
		lostsync = 0;
	    }
	} /* done running decoder */
	/* repeat ... */
    }
}
